
(* Inductive graphs, Martin Erwig syle *)

module type Graph
	=
	sig				
		type node =
			int			
		type context =
			{succ: node list; pred: node list}
		type
			graph
		val empty:
			graph
		val cons:
			node -> context -> graph -> graph 
		val decons:
			graph -> node -> (context * graph)
		val uncons:
			graph -> (node * context * graph)
	end

	
module type Indexed
	=
	sig
		type 'a indexed
		val empty : 'a indexed
		val lookup: 'a indexed -> int -> 'a 
		val insert: 'a indexed -> int -> 'a -> 'a indexed
		val remove: 'a indexed -> int -> 'a indexed
		val choose: 'a indexed -> int
		val update: 'a indexed -> int list -> ('a->'a) -> 'a indexed
	end

	
module Trie
	:   Indexed 
	=
	struct
		type 'a indexed =
			| Empty
			| Fork of 'a indexed * 'a option * 'a indexed
		let empty =
			Empty  
		let rec lookup a i =
			match a with
			| Empty ->
				failwith "Trie.lookup"
			| Fork (l,x,r) -> 
				if i land 1 = 1 then lookup l (i/2) 
				else if i > 0 then  lookup r (i/2 -1)
				else match x with
						| None -> failwith "Trie.lookup" 
						| Some y -> y  
		let rec insert a i d =
			let l,x,r = match a with
				| Fork (l,x,r) -> l,x,r
				| Empty -> Empty,None,Empty
			in	
				if i land 1 = 1 then Fork(insert l (i/2) d,x,r) 
				else if i > 0 then Fork(l,x,insert r (i/2 -1) d)
				else Fork(l,Some d,r)
		let fork l x r =
			match l,x,r with
			| Empty,None,Empty -> Empty 
			| _ -> Fork(l,x,r)
		let rec remove a i =
			match a with
			| Empty ->
				Empty
			| Fork (l,x,r) -> 
				if i land 1 = 1 then fork (remove l (i/2)) x r 
				else if i > 0 then fork l x (remove r (i/2 -1))
				else fork l None r
		let rec choose n = function
			| Empty -> -1
			| Fork (l,Some x,r) -> n
			| Fork (l,None,r) -> 
				let ret = choose (n+n+1) l in
				if ret >= 0 then ret 
				else let ret = choose (n+n+2) r in
				if ret >= 0 then ret
				else -1
		let choose a =
			choose 0 a
		let rec odd = function
			| [] -> [] 
			| h::t -> if h land 1 = 1 then (h/2)::odd t
			          else odd t
		let rec even = function
			| [] -> [] 
			| h::t -> if h land 1 = 0 then (h/2)::even t
			          else even t
		let rec update a u f =
			match a with
			| Empty ->
				Empty
			| Fork (l,None,r) ->
				Fork(update l (odd u) f, None, update r (even u) f)
			| Fork (l,Some x,r) ->
				Fork(update l (odd u) f, Some (f x), update r (even u) f)
	end

	
module MinGraph (I:Indexed)
	: Graph
	=
	struct				
		type node =
			int			
		type context =
			{succ: node list; pred: node list}
		type graph =
			context I.indexed
		let empty =
			I.empty
		let add_succ n c =
			{c with succ = n::c.succ}
		let add_pred n c =
			{c with pred = n::c.pred}
		let cons n c g =
			let g = I.insert g n c in
			let g = I.update g c.pred (add_succ n) in
			let g = I.update g c.succ (add_pred n) in
			g
		let rec remove_edge n = function
			| [] -> []
			| h::t -> if h = n then t
						 else h::remove_edge n t
		let remove_succ n c =
			{c with succ = remove_edge n c.succ}
		let remove_pred n c =
			{c with pred = remove_edge n c.pred}
 		let decons g n =
			let c = I.lookup g n in
			let g = I.remove g n in
			let g = I.update g c.pred (remove_succ n) in
			let g = I.update g c.succ (remove_pred n) in
			c,g
 		let uncons g =
			let n = I.choose g in
			let c,g = decons n g in
			n,c,g 
	end


module GraphBasics (G:Graph)
	=
	struct
		include G
		let rec map f g =
			if g = empty then empty else 
			let n,c,g = uncons g in cons n (f c) (map f g)
		let transpose g =
			map (function {succ=s; pred=p} -> {succ=p; pred=s}) g
	end
					
								