
module type TriePolitic =
	sig
		type 'a trie =
			| Empty
			| Node of 'a non_empty_trie 
		and 'a non_empty_trie =
			{mutable left: 'a trie; mutable item: 'a option; mutable right:'a trie}
		val empty :
			'a trie
		val node :
			'a non_empty_trie -> 'a trie -> 'a option -> 'a trie -> 'a non_empty_trie
		val with_left :
			'a non_empty_trie -> 'a trie -> 'a non_empty_trie 	
		val with_item :
			'a non_empty_trie -> 'a option -> 'a non_empty_trie 	
		val with_right:
			'a non_empty_trie -> 'a trie -> 'a non_empty_trie
	end

module Immutable =
	struct
		type 'a trie =
			| Empty
			| Node of 'a non_empty_trie 
		and 'a non_empty_trie =
			{mutable left: 'a trie; mutable item: 'a option; mutable right:'a trie}
		let empty =
			Empty
		let node n l i r = 	
			{left=l; item=i; right=r}
		let with_left n l = 	
			{n with left=l}
		let with_item n i = 	
			{n with item=i}
		let with_right n r = 	
			{n with right=r}
	end

module Mutable =
	struct
		type 'a trie =
			| Empty
			| Node of 'a non_empty_trie 
		and 'a non_empty_trie =
			{mutable left: 'a trie; mutable item: 'a option; mutable right:'a trie}
		let empty =
			Empty
		let node n l i r = 	
			n.left <- l; n.item <- i; n.right <- r; n
		let with_left n l = 	
			n.left <- l; n
		let with_item n i = 	
			n.item <- i; n
		let with_right n r = 	
			n.right <- r; n
	end

	
module Trie(Politic: TriePolitic) =

	struct
	
		include Politic
			
		let rec lookup a i =
			match a with
			| Empty ->
				failwith "Trie.lookup"
			| Node n -> 
				if i land 1 = 1 then lookup n.left (i/2) 
				else if i > 0 then  lookup n.right (i/2 -1)
				else match n.item with
						| None -> failwith "Trie.lookup" 
						| Some y -> y  			
						
		let rec insert a i d =
			Node
			( match a with
			| Empty -> {left=Empty; item=Some d; right=Empty}
			| Node n -> 
				if i land 1 = 1 then with_left n (insert n.left (i/2) d)
				else if i > 0 then with_right n (insert n.right (i/2 -1) d)
				else with_item n (Some d))

		let rec fold2 ta tb f =
			match ta with
			| Empty  -> f#empty_left tb
			| Node a ->
			match tb with
			| Empty  -> f#empty_right ta
			| Node b -> f#nodes a.item b.item
							(fold2 a.left b.left f)
							(fold2 a.right b.right f)	
							
		let option_or sa sb =
			match sa,sb with
			| None,None -> None
			| _,_-> sb
							
		let option_and sa sb =
			match sa,sb with
			| Some _,Some _ -> sb
			| _,_-> None
			
		let option_andnot sa sb =
			match sa,sb with
			| Some _,None -> sa
			| _,_-> None
							
		let option_xnor sa sb =
			match sa,sb with
			| Some a,Some b -> a=b
			| None,_-> true
			| _ -> false
												
		let union ta tb =
			fold2 ta tb ( 
				object
					method empty_left r = r
					method empty_right l = l
					method nodes i j l r = Node
							 {left=l; item=option_or i j; right=r}
				end )

		let intersection ta tb =
			fold2 ta tb (
				object
					method empty_left r = empty
					method empty_right l = empty
					method nodes i j l r = Node
							 {left=l; item=option_and i j; right=r}
				end )
				
		let difference ta tb =
			fold2 ta tb (
				object
					method empty_left r = empty
					method empty_right l = l
					method nodes i j l r = Node
							 {left=l; item=option_andnot i j; right=r}
				end )
	
		let subset ta tb =
			fold2 ta tb (
				object
					method empty_left r = true
					method empty_right l = false
					method nodes i j l r = l && option_xnor i j && r
				end )
				
		let equal ta tb =
			subset ta tb && subset tb ta
				
		(* utilities for Martin Erwig inductives graphs *)
		
		let rec odd = function
			| [] -> [] 
			| h::t -> if h land 1 = 1 then (h/2)::odd t
			          else odd t
			          
		let rec even = function
			| [] -> [] 
			| h::t -> if h land 1 = 0 then (h/2)::even t
			          else even t
			          
		let rec update a u f =
			match a with
			| Empty  -> Empty
			| Node n -> Node
				( match u with
				| []  -> n 
				| [0] ->
					with_item n
					( match n.item with
					| None -> None
					| Some x -> Some (f x) )			
				| _ ->			          
					node n
				  	( update n.left (odd u) f)
					( match n.item with
					| None -> None
					| Some x -> Some (f x) )
					( update n.right (even u) f)
				)				

		(* ensures any empty trie = Empty *)
		let smart_node n l x r =
			match l,x,r with
			| Empty,None,Empty -> Empty
			| _ -> Node (node n l x r)
								
		let rec remove a i =
			match a with
			| Empty ->
				Empty
			| Node n ->
				if i land 1 = 1 then
					smart_node n (remove n.left (i/2)) n.item n.right 
				else if i > 0 then
					smart_node n n.left n.item (remove n.right (i/2 -1))
				else
					smart_node n n.left None n.right
					
		let rec choose i = function
			| Empty -> -1
			| Node {item=Some x} -> i
			| Node n -> 
				let j = choose (i+i+1) n.left in
				if  j >= 0 then j 
				else choose (i+i+2) n.right
		let choose a =
			choose 0 a
								
	end
	