
(* Mutable/immutable totally ordered Sets *)

module type SetConstructors =
	sig
		type 'a set =
			'a non_empty_set option
		and 'a non_empty_set =
			private
			{mutable left: 'a set; item: 'a; mutable right: 'a set}
		val empty : 'a set
		val make : 'a set -> 'a -> 'a set -> 'a non_empty_set
		val with_left  : 'a non_empty_set -> 'a set -> 'a non_empty_set
		val with_right : 'a non_empty_set -> 'a set -> 'a non_empty_set
	end

module PureSet : SetConstructors =
	struct
		type 'a set =
			'a non_empty_set option
		and 'a non_empty_set =
			{mutable left: 'a set; item: 'a; mutable right: 'a set}
		let empty = None
		let make l x r = {left=l; item=x; right=r}
		let with_left  s l = {s with left=l}
		let with_right s r = {s with right=r}
	end

module MutableSet : SetConstructors =
	struct
		type 'a set =
			'a non_empty_set option
		and 'a non_empty_set =
			{mutable left: 'a set; item: 'a; mutable right: 'a set}
		let empty = None
		let make l x r = {left=l; item=x; right=r}
		let with_left  s l = s.left  <- l; s
		let with_right s r = s.right <- r; s
	end

module SetRecursors(S:SetConstructors) =
	struct
		include S
		(* functor *)
		let rec map s f =
			match s with
				| None ->
						None
				| Some {left=l; item=x; right=r} ->
						Some (make (map l f) (f x) (map r f))
		(* strict catamorphism *)
		type ('a,'b) fold =
			<empty: 'a; node: 'a -> 'b -> 'a -> 'a>
		let rec fold s (case:('a,'b) fold) =
			match s with
			| None -> case#empty
			| Some n -> case#node (fold n.left case) n.item (fold n.right case)
		(* lazy catamorphism *)
		type ('a,'b) cata =
			<empty: 'a; node: (unit -> 'a) -> 'b -> (unit -> 'a) -> 'a>
		let rec cata s (case:('a,'b) cata) () =
			match s with
			| None -> case#empty
			| Some n -> case#node (cata n.left case) n.item (cata n.right case)
		(* strict paramorphism *)
		type ('a,'b) recu =
			<empty: 'b;
			 node: 'a non_empty_set -> 'b -> 'b -> 'b>
		let rec recu s (case:('a,'b) recu) =
			match s with
			| None ->
					case#empty
			| Some n ->
					case#node n (recu n.left case) (recu n.right case)
		(* lazy paramorphism *)
		type ('a,'b) para =
			<empty: 'b;
			 node: 'a non_empty_set -> (unit -> 'b) -> (unit -> 'b) -> 'b>
		let rec para s (case:('a,'b) para) () =
			match s with
			| None ->
					case#empty
			| Some n ->
					case#node n (para n.left case) (para n.right case)
		let para_non_empty n (case:('a,'b) para) () =
			case#node n (para n.left case) (para n.right case)
	end

module MakeSet(S:SetConstructors)
	=
	struct
		(* general *)
		include
			SetRecursors(S)
		let cardinal s =
			fold s (
				object
					method empty = 0
					method node l _ r = l + 1 + r
				end )
		let for_all cond s =
			cata s (
				object
					method empty = true
					method node l y r =
						cond y && l() && r()
				end ) ()
		(* binary tree set *)
		let set_member x s =
			cata s (
				object
					method empty = false
					method node l y r =
						if x < y then l()
						else if x > y then r()
						else true
				end ) ()
		let set_insert x s =
			para s (
				object
					method empty =
						make None x None
					method node n l r =
						let y = n.item in
						if x < y then with_left n (Some (l()))
						else if x > y then with_left n (Some (r()))
						else n
				end ) ()
		let set_min s =
			para_non_empty s (
				object
					method empty = s
					method node n l r =
						if n.left = empty then n
						else l()
				end ) ()
		let set_remove_min s =
			para_non_empty s (
				object
					method empty = empty
					method node n l r =
						if n.left = empty then n.right
						else Some (with_left n (l()))
				end ) ()
		let set_concat sa sb =
			if sa = empty then sb else
			match sb with
			| None -> sa
			| Some n ->
				let m = set_min n
				and r = set_remove_min n
				in Some (with_right (with_left m sa) r)
		let set_remove x s =
			para s (
				object
					method empty =
						empty
					method node n l r =
						let y = n.item in
						if x < y then Some (with_left n (l()))
						else if x > y then Some (with_right n (r()))
						else set_concat n.left n.right
				end ) ()
		let set_split x s =
			para s (
				object
					method empty =
						empty,false,empty
					method node n l r =
						let y = n.item in
						if x < y then
							let a,b,c = l() in
							a,b,Some (with_left n c)
						else if x > y then
							let a,b,c = r() in
							Some (with_right n a),b,c
						else
							n.left,true,n.right
				end ) ()
		let set_filter cond s =
			recu s (
				object
					method empty =
						empty
					method node n l r =
						if cond n.item then
							Some (with_right (with_left n l) r)
						else
							set_concat l r
				end )
		let set_union sa sb =
			para sa (
				object
					method empty s =
						s
					method node m l r s =
						Some (
						match s with
						| None -> m
						| Some n ->
							let a,b,c = set_split m.item s in
							with_right (with_left m (l()a)) (r()c))
				end ) () sb
		let set_intersection sa sb =
			para sa (
				object
					method empty s =
						empty
					method node m l r s =
						match s with
						| None -> empty
						| Some n ->
							let a,b,c = set_split m.item s in
							if b then
								Some (with_right (with_left m (l()a)) (r()c))
							else
								set_concat (l()a) (r()c)
				end ) () sb
		let set_difference sa sb =
			para sa (
				object
					method empty s =
						empty
					method node m l r s =
						match s with
						| None ->
							Some m
						| Some n ->
							let a,b,c = set_split m.item s in
							if b then
								set_concat (l()a) (r()c)
							else
								Some (with_right (with_left m (l()a)) (r()c))
				end ) () sb
		let set_subset sa sb =
			para sa (
				object
					method empty s =
						true
					method node m l r s =
						match s with
						| None -> false
						| Some n ->
							if m.item < n.item then
								(l () n.left) && set_member m.item n.left && (r () s)
							else if m.item > n.item then
								(l () s) && set_member m.item n.right && (r () n.right)
							else
								(l () n.left) && (r () n.right)
				end ) () sb
	end
	

module MakeHeap(S:SetConstructors)
	=
	struct
		include
			SetRecursors(S)
		let heap_remove_min h =
			para_non_empty h (
				object
					method empty =	
						None
					method node n l r =
						match n.left,n.right with
						| t,None | None,t -> t
						| Some x, Some y -> Some (
							if x.item < y.item then
								make (l()) x.item n.right
							else
								make n.left y.item (r()))
				end ) ()
	end
	